
#include <sys/queue.h>

#include <stdlib.h>
#include <string.h>

#define SPHO_ENABLE_DEBUG_PRINT

#include "spho/util.h"

#include "msph/tree.h"

struct tree_parse {
	struct msph_token_stream	*s;

	struct msph_token		curr;

	SLIST_HEAD(msph_token_l, msph_token)	head;
};

#define CURR_TOKEN(parse)	(&parse->curr)

#define T(ptr)			((struct msph_tree *)ptr)
#define CTX(parse)		((parse)->s->ctx)

static void	tree_parse_init(struct tree_parse *,
		    struct msph_token_stream *);
static int	tree_parse_next_token(struct tree_parse *);
static void	tree_parse_push_token(struct tree_parse *, struct msph_token *);


static void	free_the_tree(struct msph_tree *);

static struct msph_tree_body	*tree_parse_body(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_ident	*tree_parse_ident(struct tree_parse *,
				    struct msph_tree *);

static struct msph_tree_dir *tree_parse_tpdef(struct tree_parse *,
				struct msph_tree *);
static struct msph_tree_dir *tree_parse_assert(struct tree_parse *,
				  struct msph_tree *);
static struct msph_tree_dir *tree_parse_nomindecl(
				    struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_dir *tree_parse_membdecl(
				    struct tree_parse *,
				    struct msph_tree *);

static struct msph_tree_tpexpr	*tree_parse_tpexpr(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_true(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_false(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_name(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_applargs(struct tree_parse *,
				    struct msph_tree_ident*,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_trait(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_conj(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_disj(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_impl(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_arrow(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_box(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_forall(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_paren(struct tree_parse *,
				    struct msph_tree *);

struct msph_tree_root *
msph_tree_makeroot(struct msph_ctx *ctx)
{
	struct msph_tree_root	*root;

	if ((root = malloc(sizeof(*root))) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		return (NULL);
	}


	T(root)->parent = NULL;
	T(root)->type = MSPH_TREE_ROOT;

	root->body = NULL;

	return (root);
}


/*
 * Expect to read next token, otherwise set error and do something
 * readres: (int) result of read
 * parse: struct tree_parse * to perform read
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_READ_NEXT(readres, parse, err_info, on_err) \
	do { \
		if (((readres) = tree_parse_next_token(parse)) == -1) \
			on_err; \
		if ((readres) == 0) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_EOF, \
			    err_info); \
			on_err; \
		} \
	} while (0)

/*
 * Expect current token type, otherwise set error and do something
 * parse: struct tree_parse * to perform read
 * toktype: token type expected
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_CURR_TOKEN(parse, toktype, err_info, on_err) \
	do { \
		if (CURR_TOKEN(parse)->type != toktype) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_NOMATCH, \
			    err_info); \
			on_err; \
		} \
	} while (0)

int
msph_tree_parse(struct msph_token_stream *s, struct msph_tree_root *root)
{
	int res;
	struct tree_parse p;
	struct msph_tree_body *body;

	if (root->body != NULL || T(root)->type != MSPH_TREE_ROOT) {
		MSPH_ERR(s->ctx, MSPH_ERR_INVAL);
		return (-1);
	}

	tree_parse_init(&p, s);
	EXPECT_READ_NEXT(res, &p, MSPH_TREE_ROOT, return (-1));

	if ((body = tree_parse_body(&p, T(root))) == NULL)
		return (-1);

	root->body = body;

	return (0);
}


static void
tree_parse_init(struct tree_parse *p, struct msph_token_stream *s)
{
	p->s = s;

	memset(&p->curr, 0, sizeof(p->curr));

	SLIST_INIT(&p->head);
}

static int
tree_parse_next_token(struct tree_parse *p)
{
	int ret;
	struct msph_token *tok;
	if (! SLIST_EMPTY(&p->head)) {
		tok = SLIST_FIRST(&p->head);
		SLIST_REMOVE_HEAD(&p->head, entries);
		memcpy(&p->curr, tok, sizeof(p->curr));
		free(tok);
		ret = 1;
	} else {
		ret = (int)msph_token_stream_read(&p->curr, 1, p->s);
	}
#ifdef SPHO_ENABLE_DEBUG_PRINT
	if (ret == 1) {
		char buf[256];
		msph_token_str(buf, sizeof(buf), CURR_TOKEN(p));
		SPHO_DEBUG_PRINT("tree_parse_read_token: curr=%s\n",
		    buf);
	}
#endif
	return (ret);
}

static struct msph_tree_body *
tree_parse_body(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_dir *dir;
	struct msph_tree_body *body;
	struct msph_token *tok;

	if ((body = malloc(sizeof(*body))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(body)->type = MSPH_TREE_BODY;
	T(body)->parent = parent;
	STAILQ_INIT(&body->head);

	while ((res = tree_parse_next_token(p)) > 0) {
		dir = NULL;
		switch(CURR_TOKEN(p)->type) {
		case TOK_KW_TYPE:
			dir = tree_parse_tpdef(p, T(body));
			break;
		case TOK_KW_NOMINAL:
			dir = tree_parse_nomindecl(p, T(body));
			break;
		case TOK_KW_MEMBER:
			dir = tree_parse_membdecl(p, T(body));
			break;
		case TOK_KW_ASSERT:
			dir = tree_parse_assert(p, T(body));
			break;
		case TOK_RBRACE:
			if ((tok = msph_token_create(CTX(p), TOK_RBRACE, NULL))
			    == NULL)
				goto err;
			tree_parse_push_token(p, tok);
			goto ret;
		default:
			MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_NOMATCH,
			    MSPH_TREE_BODY);
			goto err;
		}

		if (dir == NULL)
			goto err;
	}

	if (res == -1)
		goto err;
ret:
	return (body);
err:
	free_the_tree(T(body));

	return (NULL);
}

void
tree_parse_push_token(struct tree_parse *p, struct msph_token *tok)
{
	SLIST_INSERT_HEAD(&p->head, tok, entries);
	CURR_TOKEN(p)->type = TOK_INVAL;
}


struct msph_tree_dir *
tree_parse_tpdef(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpdef *def;

	EXPECT_CURR_TOKEN(p, TOK_KW_TYPE, MSPH_TREE_TPDEF, return (NULL));

	if ((def = malloc(sizeof(*def))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(def)->type = MSPH_TREE_TPDEF;
	T(def)->parent = parent;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->id = tree_parse_ident(p, T(def))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	EXPECT_CURR_TOKEN(p, TOK_EQUALS, MSPH_TREE_TPDEF, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->tp = tree_parse_tpexpr(p, T(def))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)def);
err:
	free_the_tree(T(def));
	return (NULL);
}

struct msph_tree_ident *
tree_parse_ident(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_ident *id;

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_IDENT, return (NULL));

	if ((id = malloc(sizeof(*id))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(id)->type = MSPH_TREE_IDENT;
	T(id)->parent = parent;

	if (SPHO_STRLCPY(id->str, CURR_TOKEN(p)->data.str, sizeof(id->str))
	    >= sizeof(id->str)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TREE_TOOLONG);
		goto err;
	}

	return (id);
err:
	free_the_tree(T(id));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_tpexpr(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_tpexpr *tp;

	switch (CURR_TOKEN(p)->type) {
	case TOK_AMP:
		tp = tree_parse_conj(p, parent);
		break;
	case TOK_PIPE:
		tp = tree_parse_disj(p, parent);
		break;
	case TOK_IMPL:
		tp = tree_parse_impl(p, parent);
		break;
	case TOK_RARROW:
		tp = tree_parse_arrow(p, parent);
		break;
	case TOK_KW_BOX:
		tp = tree_parse_box(p, parent);
		break;
	case TOK_KW_FORALL:
		tp = tree_parse_forall(p, parent);
		break;
	case TOK_IDENT:
		tp = tree_parse_name(p, parent);
		break;
	case TOK_CONST_TRUE:
		tp = tree_parse_true(p, parent);
		break;
	case TOK_CONST_FALSE:
		tp = tree_parse_false(p, parent);
		break;
	case TOK_LPAREN:
		tp = tree_parse_paren(p, parent);
		break;
	case TOK_LBRACE:
		tp = tree_parse_trait(p, parent);
		break;
	default:
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_NOMATCH, MSPH_TREE_TPEXPR);
		tp = NULL;
	}

	return (tp);
}

struct msph_tree_tpexpr *
tree_parse_conj(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_conj *conj;

	EXPECT_CURR_TOKEN(p, TOK_AMP, MSPH_TREE_CONJ, return (NULL));

	if ((conj = malloc(sizeof(*conj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(conj)->type = MSPH_TREE_CONJ;
	T(conj)->parent = parent;
	conj->ltp = NULL;
	conj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);

	if ((conj->ltp = tree_parse_tpexpr(p, (struct msph_tree *)conj)) 
	    == NULL)
		goto err;

	if ((conj->rtp = tree_parse_tpexpr(p, (struct msph_tree *)conj))
	    == NULL)
		goto err;
err:
	free_the_tree((struct msph_tree *)conj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_disj(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_disj *disj;

	EXPECT_CURR_TOKEN(p, TOK_PIPE, MSPH_TREE_DISJ, return (NULL));

	if ((disj = malloc(sizeof(*disj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(disj)->type = MSPH_TREE_DISJ;
	T(disj)->parent = parent;
	disj->ltp = NULL;
	disj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);

	if ((disj->ltp = tree_parse_tpexpr(p, (struct msph_tree *)disj)) 
	    == NULL)
		goto err;

	if ((disj->rtp = tree_parse_tpexpr(p, (struct msph_tree *)disj))
	    == NULL)
		goto err;
err:
	free_the_tree((struct msph_tree *)disj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_impl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_impl *impl;

	EXPECT_CURR_TOKEN(p, TOK_PIPE, MSPH_TREE_IMPL, return (NULL));

	if ((impl = malloc(sizeof(*impl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(impl)->type = MSPH_TREE_IMPL;
	T(impl)->parent = parent;
	impl->ltp = NULL;
	impl->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);

	if ((impl->ltp = tree_parse_tpexpr(p, (struct msph_tree *)impl)) 
	    == NULL)
		goto err;

	if ((impl->rtp = tree_parse_tpexpr(p, (struct msph_tree *)impl))
	    == NULL)
		goto err;
err:
	free_the_tree((struct msph_tree *)impl);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_arrow(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_arrow *arrw;

	EXPECT_CURR_TOKEN(p, TOK_PIPE, MSPH_TREE_ARROW, return (NULL));

	if ((arrw = malloc(sizeof(*arrw))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(arrw)->type = MSPH_TREE_ARROW;
	T(arrw)->parent = parent;
	arrw->ltp = NULL;
	arrw->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);

	if ((arrw->ltp = tree_parse_tpexpr(p, (struct msph_tree *)arrw)) 
	    == NULL)
		goto err;

	if ((arrw->rtp = tree_parse_tpexpr(p, (struct msph_tree *)arrw))
	    == NULL)
		goto err;
err:
	free_the_tree((struct msph_tree *)arrw);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_box(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_box *box;

	EXPECT_CURR_TOKEN(p, TOK_KW_BOX, MSPH_TREE_BOX, return (NULL));

	if ((box = malloc(sizeof(*box))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(box)->type = MSPH_TREE_BOX;
	T(box)->parent = parent;
	box->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_BOX, goto err);
	if ((box->inner = tree_parse_tpexpr(p, T(box))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)box);
err:
	free_the_tree(T(box));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_forall(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_forall *fa;

	EXPECT_CURR_TOKEN(p, TOK_KW_FORALL, MSPH_TREE_FORALL, return (NULL));

	if ((fa = malloc(sizeof(*fa))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(fa)->type = MSPH_TREE_FORALL;
	T(fa)->parent = parent;
	fa->ident = NULL;
	fa->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->ident = tree_parse_ident(p, T(fa))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_COLON, MSPH_TREE_FORALL, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->inner = tree_parse_tpexpr(p, T(fa))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)fa);
err:
	free_the_tree(T(fa));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_name(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_ident *id;
	struct msph_tree_name *name;
	struct msph_tree_tpexpr *ret;

	ret = NULL;
	name = NULL;

	if ((id = tree_parse_ident(p, NULL)) == NULL)
		return (NULL);

	if ((res = tree_parse_next_token(p)) == 1 &&
	    CURR_TOKEN(p)->type == TOK_LBRAK) {
		if ((ret = tree_parse_applargs(p, id, parent)) == NULL)
			goto err;
		return (ret);
	} else if (res == -1) {
		goto err;
	}

	if ((name = malloc(sizeof(*name))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}
	T(name)->type = MSPH_TREE_NAME;
	T(name)->parent = parent;

	name->id = id;

	return ((struct msph_tree_tpexpr *)name);
err:
	free_the_tree(T(id));

	return (NULL);
}

/**
 * Parse argument list part [...], of type application A[...].
 *
 * @param p Parse state
 * @param id Parsed ident A (on failure, this is not freed)
 * @param parent Parent of resulting appl node
 *
 * @return tpexpr node representing A[...].
 */
struct msph_tree_tpexpr *
tree_parse_applargs(struct tree_parse *p, struct msph_tree_ident *id,
    struct msph_tree *parent)
{
	int res;
	struct msph_tree_appl *appl;
	struct msph_tree_tpexpr *tp;

	tp = NULL;

	EXPECT_CURR_TOKEN(p, TOK_LBRAK, MSPH_TREE_APPL, return (NULL));

	if ((appl = malloc(sizeof(*appl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(appl)->type = MSPH_TREE_APPL;
	T(appl)->parent = parent;
	appl->id = NULL;
	STAILQ_INIT(&appl->head);

	/* parse the argument list */
	while ((res = tree_parse_next_token(p)) == 1) {
		/* if ']' we have reached end of arguments */
		if (CURR_TOKEN(p)->type == TOK_RBRAK)
			break;

		/* if not start of argument list, check for comma and step
		 * tokenizer */
		if (tp != NULL) {
			EXPECT_CURR_TOKEN(p, TOK_COMMA, MSPH_TREE_APPL,
			    goto err);
			EXPECT_READ_NEXT(res, p, MSPH_TREE_APPL, goto err);
		}

		if ((tp = tree_parse_tpexpr(p, T(appl))) == NULL)
			goto err;

		STAILQ_INSERT_TAIL(&appl->head, tp, entries);
	}

	/* system error */
	if (res == -1)
		goto err;

	/* did not read final ']' because of EOF */
	if (res == 0) {
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_EOF, MSPH_TREE_APPL);
		goto err;
	}

	/* finally, appl takes ownership of id */
	appl->id = id;

	return ((struct msph_tree_tpexpr *)appl);
err:
	free_the_tree(T(appl));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_true(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *t;

	EXPECT_CURR_TOKEN(p, TOK_CONST_TRUE, MSPH_TREE_TRUE, return (NULL));

	if ((t = malloc(sizeof(*t))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(t)->type = MSPH_TREE_TRUE;
	T(t)->parent = parent;

	return ((struct msph_tree_tpexpr *)t);
}

struct msph_tree_tpexpr *
tree_parse_false(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *f;

	EXPECT_CURR_TOKEN(p, TOK_CONST_FALSE, MSPH_TREE_FALSE, return (NULL));

	if ((f = malloc(sizeof(*f))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(f)->type = MSPH_TREE_FALSE;
	T(f)->parent = parent;

	return ((struct msph_tree_tpexpr *)f);
}

struct msph_tree_tpexpr *
tree_parse_paren(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_paren *par;

	EXPECT_CURR_TOKEN(p, TOK_LPAREN, MSPH_TREE_FALSE, return (NULL));

	if ((par = malloc(sizeof(*par))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(par)->type = MSPH_TREE_PAREN;
	T(par)->parent = parent;
	par->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);

	if ((par->inner = tree_parse_tpexpr(p, T(par))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RPAREN, MSPH_TREE_PAREN, goto err);

	return ((struct msph_tree_tpexpr *)par);
err:
	free_the_tree(T(par));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_trait(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_trait *trait;

	EXPECT_CURR_TOKEN(p, TOK_LBRACE, MSPH_TREE_TRAIT, return (NULL));

	if ((trait = malloc(sizeof(*trait))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	T(trait)->type = MSPH_TREE_TRAIT;
	T(trait)->parent = parent;
	trait->body = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TRAIT, goto err);

	if ((trait->body = tree_parse_body(p, T(trait))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TRAIT, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RBRACE, MSPH_TREE_TRAIT, goto err);

	return ((struct msph_tree_tpexpr *)trait);
err:
	free_the_tree(T(trait));

	return (NULL);
}

static void
free_the_tree(struct msph_tree *tree)
{
	/* no one can free the tree :´( */
	SPHO_POSTCOND((void *)tree == (void *)free);
}


